/*
  DXC Framework (http://dx-competition.org/)
  Copyright 2008 David Garcia

  The DXC Framework is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation, either version 3 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>
*/

#include <iostream>
#include <exception>
#include <typeinfo>
#include <algorithm>

#include <DxcDataTypes.hpp>
#include <Connection/ConnectorFactory.hpp>

#include <DAConfig.hpp>
#include <ScenarioLoader.hpp>
#include <ProcessFactory.hpp>
#include <FileLister.hpp>
#include <DxcConfig.hpp>

#ifdef _WIN32
#include <Windows.h>
#include <Pthreads-w32/pthread.h>
#define pthreadsInit pthread_win32_process_attach_np
#define pthreadsStop pthread_win32_process_detach_np
#else
#include <unistd.h>
#include <pthread.h>
#define pthreadsInit void
#define pthreadsStop void
#endif

using namespace std;
using namespace Dxc;

static bool abbrev = false;
SLCallback::SLCallback(ScenarioTracker *scenarioTracker) :

    scenarioTracker_(scenarioTracker)
{
};

void SLCallback::processData(const DxcData * msg)
{
    if (string(typeid(*msg).name()) ==
            string(typeid(ScenarioStatusData).name()))
    {
        const ScenarioStatusData *stat;

        stat = dynamic_cast<const ScenarioStatusData *>(msg);

        string status = stat->getStatus();

        if (status == ScenarioStatusData::SDS_AWAITING_DA)
        {
            scenarioTracker_->updateStatus(ScenarioTracker::ST_AWAITING_DA);
        }
        else if (status == ScenarioStatusData::STARTED)
        {
            scenarioTracker_->updateStatus(ScenarioTracker::ST_STARTED);
        }
        else if (status == ScenarioStatusData::SDS_ENDED)
        {
            scenarioTracker_->updateStatus(ScenarioTracker::ST_ENDED);
        }
    }
    else if (typeid(msg).name() == typeid(ErrorData).name())
    {
        scenarioTracker_->updateStatus(ScenarioTracker::ST_ERROR);
        cout << msg;
    }
};

vector<DAConfig> getDAConfigs(const vector<string>& daDirs)
{
    vector<DAConfig> ret;
    vector<string>::const_iterator i = daDirs.begin();
    string algsDir;

    try
    {
        DxcConfig* cfg = DxcConfig::getInstance();
        algsDir = cfg->get("ALGS_DIR");

        for (; i != daDirs.end(); i++)
        {
            ret.push_back(DAConfig(algsDir + DIR_DELIM + *i + DIR_DELIM + "config.xml"));
        }
    }
    catch (DxcException ex)
    {
        cout << "SL: ERROR: ScenarioLoader unable to load config for DA in "
             << endl << algsDir << DIR_DELIM << *i << endl;
        cout << "Exception: " << ex.what() << endl;
    }

    return ret;
}

bool runScenario(const DAConfig &daConf,
                 const string &scenarioFile, const string &systemName)
{
    ScenarioTracker *strack(0);
    SLCallback *slcb(0);
    DxcConnector *slc(0);
    ProcessLoader *sds(0);
    ProcessLoader *sr(0);
    ProcessLoader *da(0);
    ProcessLoader *orc(0);

    try
    {
        DxcConfig* cfg = DxcConfig::getInstance();

        string scenDir = cfg->get("SCENARIOS_DIR");
        string binDir = cfg->get("BIN_DIR");
        string resultsDir = cfg->get("RESULTS_DIR");

        string scenario = "\"" + scenDir + DIR_DELIM +
                          systemName + DIR_DELIM +
                          scenarioFile + "\"";

        string srCmd = "\"" + daConf.getName() + "\" " +
                       "\"" + daConf.getVersion() + "\" " +
                       "\"" + resultsDir + DIR_DELIM +
                       daConf.getName() +
                       " - " + systemName +
                       " - " + scenarioFile + "\"";

        string sdsCmd;
        if (abbrev)
            sdsCmd = "-a ";
        sdsCmd.append("-s -v " + systemName + " " + scenario);
        sds =
            ProcessFactory::createProcessLoader(
                "SDS",
                binDir,
                sdsCmd);

        sr =
            ProcessFactory::createProcessLoader(
                "ScenarioRecorder",
                binDir,
                srCmd);

        cout << "SL: INFO: Loading Diagnoser: " << daConf.getName() << endl;
        da = ProcessFactory::createDAProcessLoader(daConf, systemName);

        strack = new ScenarioTracker();
        slcb = new SLCallback(strack);
        slc = ConnectorFactory::getScenarioLoaderConnector(slcb);

        sds->start();
        msleep(1000);
        sr->start();

        if (strack->getStatus() != ScenarioTracker::ST_AWAITING_DA)
        {
            strack->waitForUpdate(cfg->getLong("SDS_SR_TIMEOUT_MS"));
            if (strack->getStatus() != ScenarioTracker::ST_AWAITING_DA)
            {
                throw DxcException("Timed out waiting for SDS and SR.");
            }
        }

        slc->initConnections();


        if (cfg->get("RUN_ORACLE") == "TRUE")
        {
            orc =
                ProcessFactory::createProcessLoader(
                    "Oracle",
                    binDir,
                    "[args]");
            orc->start();
        }

        msleep(1000);
        da->start();

        if (strack->getStatus() != ScenarioTracker::ST_STARTED &&
                strack->getStatus() != ScenarioTracker::ST_ENDED)
        {
            long l = cfg->getLong("DA_TIMEOUT_MS");

            strack->waitForUpdate(cfg->getLong("DA_TIMEOUT_MS"));
            if (strack->getStatus() != ScenarioTracker::ST_STARTED &&
                    strack->getStatus() != ScenarioTracker::ST_ENDED)
            {
                throw DxcException("Timed out waiting for scenario start (check DA -> SDS connection).");
            }
        }
        if (strack->getStatus() != ScenarioTracker::ST_ENDED)
        {
            strack->waitForUpdate(cfg->getLong("SCN_END_TIMEOUT_MS"));
        }
        if (strack->getStatus() != ScenarioTracker::ST_ENDED)
        {
            if (strack->getStatus() == ScenarioTracker::ST_ERROR)
            {
                throw DxcException("Received error message. Terminating.");
            }

            throw DxcException("Timed out waiting for scenario end.");
        }

        if (!da->timedWait(cfg->getLong("POST_SCN_TIME_MS")))
        {
            cout << "SL: WARNING: DA has not exited. Terminating." << endl;

            da->terminate();
        }

        ScenarioStatusData ssd(ScenarioStatusData::DA_ENDED);
        slc->sendMessage(&ssd);

        ProfilingData profilingData(da->getCpuTime(),
                                    da->getPeakMemoryUsage());
        slc->sendMessage(&profilingData);

        delete slc;
        delete slcb;
        // wait for apps to exit gracefully
        msleep(1000);
        delete sds;
        delete sr;
        delete orc;
        delete da;
        delete strack;
    }
    catch (DxcException ex)
    {
        string err("SL: ERROR: ");
        err += ex.what();
        cerr << err << endl;

        cout << "SL: ERROR: Exception. Terminating DA." << endl;
        try
        {
            da->terminate();
        }
        catch (...)
        {
        }

        if (slc)
        {
            ErrorData ed(DxcData::currentTime(), err);
            ScenarioStatusData ssd(ScenarioStatusData::DA_ENDED);
            slc->sendMessage(&ed);
            slc->sendMessage(&ssd);
            try
            {
                ProfilingData profilingData(da->getCpuTime(),
                                            da->getPeakMemoryUsage());
                slc->sendMessage(&profilingData);
            }
            catch (...)
            {
            }
        }

        delete slc;
        delete slcb;
        // wait for apps to exit gracefully
        msleep(1000);
        delete sds;
        delete sr;
        delete da;
        delete strack;
    }

    return true;
}

bool runExperiment(const DAConfig& daConf)
{
    vector<string> systems = daConf.getSupportedSystems();
    vector<string>::const_iterator i;

    DxcConfig* cfg = DxcConfig::getInstance();
    string scenDir = cfg->get("SCENARIOS_DIR");

    for (i = systems.begin(); i != systems.end(); i++)
    {
        vector<string> scenarios;

        try
        {
            scenarios = FileLister::getFiles(scenDir + DIR_DELIM + *i + DIR_DELIM);
            sort(scenarios.begin(), scenarios.end());
        }
        catch (...)
        {
            cerr << "SL: ERROR: Unable to get scenarios for system " << *i << endl;
            continue;
        }

        vector<string>::const_iterator j = scenarios.begin();

        for (; j  != scenarios.end(); j++)
        {
            runScenario(daConf, *j, *i);
            msleep(1000);
        }
    }
    return true;
}

int main(int argc, char* argv[])
{
    pthreadsInit();
    if (argc > 1)
        abbrev = true;

    try
    {
        DxcConfig *cfg = DxcConfig::getInstance();
        string algsDir = cfg->get("ALGS_DIR");

        cout << "SL: INFO: Iterating through DAs in: " << endl << algsDir << endl;
        vector<string> daDirs = FileLister::getDirectories(algsDir);
        vector<DAConfig> daCfgs = getDAConfigs(daDirs);

        vector<DAConfig>::const_iterator i;

        bool fg = false;
        for (i = daCfgs.begin(); i != daCfgs.end(); i++)
        {
            if (!runExperiment(*i))
            {
                fg = true;
                break;
            }
        }

        if (!fg)
        {
            cout << "SL: INFO: Exiting..." << endl;
        }

        delete cfg;

        /*      msleep(500); */
    }
    catch (DxcException ex)
    {
        cerr << "SL: ERROR: " << ex.what() << endl;
    }

    pthreadsStop();
}
